package lcsb.mapviewer.services.impl;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.commands.ColorExtractor;
import lcsb.mapviewer.common.Configuration;
import lcsb.mapviewer.common.ObjectUtils;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.common.geometry.ColorParser;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.log.LogType;
import lcsb.mapviewer.model.user.BasicPrivilege;
import lcsb.mapviewer.model.user.ConfigurationElementType;
import lcsb.mapviewer.model.user.ObjectPrivilege;
import lcsb.mapviewer.model.user.PrivilegeType;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.persist.dao.ProjectDao;
import lcsb.mapviewer.persist.dao.user.PrivilegeDao;
import lcsb.mapviewer.persist.dao.user.UserDao;
import lcsb.mapviewer.services.AuthenticationTokenExpireException;
import lcsb.mapviewer.services.InvalidTokenException;
import lcsb.mapviewer.services.SecurityException;
import lcsb.mapviewer.services.interfaces.IConfigurationService;
import lcsb.mapviewer.services.interfaces.ILogService;
import lcsb.mapviewer.services.interfaces.ILogService.LogParams;
import lcsb.mapviewer.services.interfaces.IUserService;
import lcsb.mapviewer.services.view.AuthenticationToken;
import lcsb.mapviewer.services.view.PrivilegeView;
import lcsb.mapviewer.services.view.UserView;
import lcsb.mapviewer.services.view.UserView.UserProjectPrivilegeView;
import lcsb.mapviewer.services.view.UserViewFactory;

/**
 * Implementation of the service that manages users.
 * 
 * @author Piotr Gawron
 * 
 */
@Transactional(value = "txManager")
public class UserService implements IUserService {

	private static Map<String, AuthenticationToken>	authenticationTokens = new HashMap<>();

	private static Map<AuthenticationToken, User>		authenticatedUsers	 = new HashMap<>();

	/**
	 * Default class logger.
	 */
	private static Logger														logger							 = Logger.getLogger(UserService.class);

	/**
	 * Data access object for users.
	 */
	@Autowired
	private UserDao																	userDao;

	/**
	 * Data access object for projects.
	 */
	@Autowired
	private ProjectDao															projectDao;

	/**
	 * Data access object for privileges.
	 */
	@Autowired
	private PrivilegeDao														privilegeDao;

	/**
	 * Factory object for {@link UserView} elements.
	 */
	@Autowired
	private UserViewFactory													userViewFactory;

	/**
	 * Service that provides password encoding.
	 */
	@Autowired
	private PasswordEncoder													passwordEncoder;

	/**
	 * Service used for logging.
	 */
	@Autowired
	private ILogService															logService;

	/**
	 * Service used for accessing confguration parameters.
	 */
	@Autowired
	private IConfigurationService										configurationService;

	@Override
	public AuthenticationToken login(String login, String password) {
		User user = userDao.getUserByLoginAndPassword(login, password);
		if (Configuration.ANONYMOUS_LOGIN.equals(login)) {
			user = getUserByLogin(Configuration.ANONYMOUS_LOGIN);
		}
		if (user != null) {
			int count = 0;
			synchronized (authenticationTokens) {
				count = authenticationTokens.size();
			}

			if (count > 1000) {
				clearAuthenticationTokens();
			}
			AuthenticationToken authenticationToken = new AuthenticationToken();
			synchronized (authenticationTokens) {
				authenticationTokens.put(authenticationToken.getId(), authenticationToken);
				authenticatedUsers.put(authenticationToken, user);
			}
			return authenticationToken;
		} else {
			return null;
		}
	}

	private void clearAuthenticationTokens() {
		synchronized (authenticationTokens) {
			List<String> toRemove = new ArrayList<>();
			for (AuthenticationToken token : authenticationTokens.values()) {
				if (token.getExpires().before(Calendar.getInstance())) {
					toRemove.add(token.getId());
				}
			}
		}
	}

	@Override
	public boolean userHasPrivilege(User user, PrivilegeType type) {
		return getUserPrivilegeLevel(user, type) > 0;
	}

	@Override
	public boolean userHasPrivilege(User user, PrivilegeType type, Object object) {
		return getUserPrivilegeLevel(user, type, object) > 0;
	}

	@Override
	public void setUserPrivilege(User user, BasicPrivilege privilege) {
		privilege.setUser(user);
		BasicPrivilege oldPrivilege = null;
		for (BasicPrivilege privilegeIter : user.getPrivileges()) {
			if (privilegeIter.equalsPrivilege(privilege)) {
				oldPrivilege = privilegeIter;
			}
		}
		if (oldPrivilege != null) {
			user.getPrivileges().remove(oldPrivilege);
			oldPrivilege.setUser(null);
		}
		user.getPrivileges().add(privilege);
		updateUser(user);
		userDao.flush();
	}

	@Override
	public void addUser(User user) {
		userDao.add(user);
		LogParams params = new LogParams().description("User " + user.getLogin() + " created.").type(LogType.USER_CREATED).object(user);
		logService.log(params);
	}

	@Override
	public void updateUser(User user) {
		userDao.update(user);
	}

	@Override
	public void deleteUser(User user) {
		userDao.delete(user);
		LogParams params = new LogParams().description("User " + user.getLogin() + " removed.").type(LogType.USER_CREATED).object(user);
		logService.log(params);
	}

	@Override
	public User getUserById(int id) {
		User result = userDao.getById(id);
		if (result != null) {
			userDao.refresh(result);
		}
		return result;
	}

	@Override
	public User getUserByLogin(String login) {
		User result = userDao.getUserByLogin(login);
		if (result != null) {
			userDao.refresh(result);
		}
		return result;
	}

	@Override
	public List<UserView> getAllUserRows() {
		List<Project> projects = projectDao.getAll();

		List<UserView> result = new ArrayList<UserView>();
		List<User> fullList = userDao.getAll();
		for (User user : fullList) {
			result.add(userViewFactory.create(user, projects));
		}

		return result;
	}

	@Override
	public UserView getUserRow(User user) {
		List<Project> projects = projectDao.getAll();
		return userViewFactory.create(user, projects);
	}

	@Override
	public void updateUser(UserView userRow) {
		User user = null;
		if (userRow.getIdObject() == null || userRow.getIdObject() == 0) {
			user = new User();
			user.setLogin(userRow.getLogin());
		} else {
			user = getUserById(userRow.getIdObject());
		}
		if (userRow.getPassword() != null && !userRow.getPassword().trim().equals("")) {
			user.setCryptedPassword(passwordEncoder.encode(userRow.getPassword()));
		}

		user.setEmail(userRow.getEmail());
		user.setName(userRow.getName());
		user.setSurname(userRow.getSurname());
		if (user.getId() == null) {
			addUser(user);
		} else {
			updateUser(user);
		}

		userRow.setIdObject(user.getId());

		for (PrivilegeView pRow : userRow.getBasicPrivileges()) {
			if (pRow.getNumeric()) {
				setUserPrivilege(user, new BasicPrivilege(pRow.getLevel(), pRow.getType(), user));
			} else if (pRow.getSelected()) {
				setUserPrivilege(user, new BasicPrivilege(1, pRow.getType(), user));
			} else {
				setUserPrivilege(user, new BasicPrivilege(0, pRow.getType(), user));
			}
		}
		for (UserProjectPrivilegeView upRow : userRow.getProjectPrivileges()) {
			for (PrivilegeView pRow : upRow.getProjectPrivileges()) {
				Project project = projectDao.getById(upRow.getIdObject());
				if (pRow.getNumeric()) {
					setUserPrivilege(user, new ObjectPrivilege(project, pRow.getLevel(), pRow.getType(), user));
				} else if (pRow.getSelected()) {
					setUserPrivilege(user, new ObjectPrivilege(project, 1, pRow.getType(), user));
				} else {
					setUserPrivilege(user, new ObjectPrivilege(project, 0, pRow.getType(), user));
				}
			}
		}
	}

	@Override
	public void deleteUser(UserView selectedUser) {
		if (selectedUser.getIdObject() != null && selectedUser.getIdObject() != 0) {
			deleteUser(getUserById(selectedUser.getIdObject()));
		}
	}

	@Override
	public void dropPrivilegesForObjectType(PrivilegeType type, int id) {
		// this will be slow when number of user will increase (we fetch all
		// users and drop privileges one by one)
		List<User> users = userDao.getAll();
		for (User user : users) {
			List<BasicPrivilege> toRemove = new ArrayList<BasicPrivilege>();
			for (BasicPrivilege privilege : user.getPrivileges()) {
				if (privilege.getType().equals(type) && privilege instanceof ObjectPrivilege && ((ObjectPrivilege) privilege).getIdObject() == id) {
					toRemove.add(privilege);
				}
			}
			if (toRemove.size() > 0) {
				user.getPrivileges().removeAll(toRemove);
				userDao.update(user);
			}
		}
	}

	@Override
	public int getUserPrivilegeLevel(User user, PrivilegeType type) {
		if (type.getPrivilegeClassType() != BasicPrivilege.class) {
			throw new InvalidArgumentException("This privilege requires additional information");
		}
		for (BasicPrivilege privilege : user.getPrivileges()) {
			if (privilege.getType().equals(type)) {
				return privilege.getLevel();
			}
		}
		return 0;
	}

	@Override
	public int getUserPrivilegeLevel(User user, PrivilegeType type, Object object) {
		if (type.getPrivilegeClassType() != ObjectPrivilege.class) {
			throw new InvalidArgumentException("This privilege doesn't accept object parameter");
		}
		if (object == null) {
			throw new InvalidArgumentException("Object cannot be null");
		}
		if (!type.getPrivilegeObjectType().isAssignableFrom(object.getClass())) {
			throw new InvalidArgumentException(
					"This privilege accept only " + type.getPrivilegeObjectType() + " objects parameter, but " + object.getClass() + " class found.");
		}
		Integer id = null;
		try {
			id = ObjectUtils.getIdOfObject(object);
		} catch (Exception e) {
			logger.error(e.getMessage(), e);
			throw new InvalidArgumentException("Internal server error. Problem with accessing id of the parameter object");
		}
		if (id == null) {
			throw new InvalidArgumentException("Parameter object has null id value");
		}
		if (user == null) {
			throw new InvalidArgumentException("User cannot be null");
		}

		// refresh user from db
		if (user.getId() != null) {
			user = userDao.getById(user.getId());
		}
		for (BasicPrivilege privilege : user.getPrivileges()) {
			if (privilege.getClass() == ObjectPrivilege.class) {
				ObjectPrivilege oPrivilege = (ObjectPrivilege) privilege;
				if (oPrivilege.getType().equals(type) && oPrivilege.getIdObject().equals(id)) {
					return privilege.getLevel();
				}
			}
		}
		return -1;
	}

	@Override
	public void addUser(UserView userRow) {
		updateUser(userRow);
	}

	@Override
	public void setUserPrivilege(User user, PrivilegeType type) {
		BasicPrivilege privilege = new BasicPrivilege(1, type, user);

		BasicPrivilege oldPrivilege = null;
		for (BasicPrivilege privilegeIter : user.getPrivileges()) {
			if (privilegeIter.getType().equals(type)) {
				oldPrivilege = privilegeIter;
			}
		}
		if (oldPrivilege != null) {
			user.getPrivileges().remove(oldPrivilege);
			oldPrivilege.setUser(null);
		}
		user.getPrivileges().add(privilege);
		updateUser(user);
		userDao.flush();

	}

	/**
	 * @return the userDao
	 * @see #userDao
	 */
	public UserDao getUserDao() {
		return userDao;
	}

	/**
	 * @param userDao
	 *          the userDao to set
	 * @see #userDao
	 */
	public void setUserDao(UserDao userDao) {
		this.userDao = userDao;
	}

	/**
	 * @return the projectDao
	 * @see #projectDao
	 */
	public ProjectDao getProjectDao() {
		return projectDao;
	}

	/**
	 * @param projectDao
	 *          the projectDao to set
	 * @see #projectDao
	 */
	public void setProjectDao(ProjectDao projectDao) {
		this.projectDao = projectDao;
	}

	/**
	 * @return the privilegeDao
	 * @see #privilegeDao
	 */
	public PrivilegeDao getPrivilegeDao() {
		return privilegeDao;
	}

	/**
	 * @param privilegeDao
	 *          the privilegeDao to set
	 * @see #privilegeDao
	 */
	public void setPrivilegeDao(PrivilegeDao privilegeDao) {
		this.privilegeDao = privilegeDao;
	}

	/**
	 * @return the passwordEncoder
	 * @see #passwordEncoder
	 */
	public PasswordEncoder getPasswordEncoder() {
		return passwordEncoder;
	}

	/**
	 * @param passwordEncoder
	 *          the passwordEncoder to set
	 * @see #passwordEncoder
	 */
	public void setPasswordEncoder(PasswordEncoder passwordEncoder) {
		this.passwordEncoder = passwordEncoder;
	}

	@Override
	public UserView createEmptyUserRow() {
		return userViewFactory.create(null);
	}

	@Override
	public void updateUsers(Collection<UserView> users) {
		for (UserView user : users) {
			updateUser(user);
		}
	}

	/**
	 * @param password
	 *          input password
	 * @return encoded password
	 */
	@Override
	public String encodePassword(String password) {
		return passwordEncoder.encode(password);
	}

	@Override
	public User getUserByNameSurname(String nameSurnameString) {
		return userDao.getUserByNameSurname(nameSurnameString);
	}

	@Override
	public ColorExtractor getColorExtractorForUser(User loggedUser) {
		Color colorMin = null;
		Color colorMax = null;
		if (loggedUser != null) {
			User dbUser = getUserById(loggedUser.getId());
			if (dbUser != null) {
				colorMin = dbUser.getMinColor();
				colorMax = dbUser.getMaxColor();
			}
		}
		ColorParser parser = new ColorParser();

		if (colorMin == null) {
			colorMin = parser.parse(configurationService.getConfigurationValue(ConfigurationElementType.MIN_COLOR_VAL));
		}
		if (colorMax == null) {
			colorMax = parser.parse(configurationService.getConfigurationValue(ConfigurationElementType.MAX_COLOR_VAL));
		}
		return new ColorExtractor(colorMin, colorMax);
	}

	/**
	 * @return the configurationService
	 * @see #configurationService
	 */
	public IConfigurationService getConfigurationService() {
		return configurationService;
	}

	/**
	 * @param configurationService
	 *          the configurationService to set
	 * @see #configurationService
	 */
	public void setConfigurationService(IConfigurationService configurationService) {
		this.configurationService = configurationService;
	}

	@Override
	public User getUserByToken(AuthenticationToken token) {
		User result = null;
		if (Calendar.getInstance().before(token.getExpires())) {
			synchronized (authenticationTokens) {
				result = authenticatedUsers.get(token);
			}
		}
		return result;
	}

	@Override
	public User getUserByToken(String tokenString) throws SecurityException {
		return getUserByToken(getToken(tokenString));
	}

	@Override
	public AuthenticationToken getToken(String tokenString) throws SecurityException {
		AuthenticationToken result = null;
		synchronized (authenticationTokens) {
			result = authenticationTokens.get(tokenString);
		}
		if (result == null) {
			throw new InvalidTokenException("Token string is invalid");
		}
		if (result.getExpires().before(Calendar.getInstance())) {
			logout(result);
			throw new AuthenticationTokenExpireException("Token validity expired");
		}
		return result;
	}

	@Override
	public void logout(AuthenticationToken token) {
		synchronized (authenticationTokens) {
			authenticationTokens.remove(token.getId());
			authenticatedUsers.remove(token);
		}
	}

	@Override
	public boolean userHasPrivilege(AuthenticationToken token, PrivilegeType type, Object object) {
		return userHasPrivilege(getUserByToken(token), type, object);
	}

	@Override
	public void logout(String tokenString) throws SecurityException {
		AuthenticationToken token = getToken(tokenString);
		logout(token);
	}

	@Override
	public boolean userHasPrivilege(AuthenticationToken token, PrivilegeType type) {
		return userHasPrivilege(getUserByToken(token), type);
	}
}
