package lcsb.mapviewer.services.interfaces;

import java.awt.geom.Point2D;
import java.util.List;

import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.map.Comment;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.services.UserAccessException;
import lcsb.mapviewer.services.search.ElementIdentifierDetails;
import lcsb.mapviewer.services.search.comment.FullCommentView;
import lcsb.mapviewer.services.search.data.ElementIdentifier;
import lcsb.mapviewer.services.view.AuthenticationToken;
import lcsb.mapviewer.services.view.CommentView;

/**
 * Service responsible for comments functionality.
 * 
 * @author Piotr Gawron
 * 
 */
public interface ICommentService {
	/**
	 * Adds comment to the map.
	 * 
	 * @param pinned
	 *          parameter that defines if comment should be visible on the map.
	 * @param object
	 *          determines which object is commented - it could be null; in this
	 *          case it means that a comment is general one.
	 * @param name
	 *          name of person that comments
	 * @param email
	 *          email of the person that comments
	 * @param content
	 *          content of the comment
	 * @param model
	 *          which map we comment on
	 * @param submodel
	 *          on which submodel we comment on
	 * @param coordinates
	 *          where exactly the comment should be placed on
	 * 
	 * @return comment object created based on the data provided as parameters
	 */
	Comment addComment(String name, String email, String content, Model model, Point2D coordinates, Object object, boolean pinned, Model submodel);

	/**
	 * This method remove comment. Comment is not removed from the system, only
	 * 'deleted' flag is set.
	 * 
	 * @param loggedUser
	 *          user that wants to remove the comment
	 * @param commentId
	 *          identifier of the comment that user wants to remove
	 * @param reason
	 *          why user wants to remove the comment
	 */
	void deleteComment(User loggedUser, String commentId, String reason);

	/**
	 * Method returns all comments for a given map.
	 * 
	 * @param model
	 *          given map
	 * @return list of the comments to the map
	 * @throws UserAccessException 
	 */
	List<CommentView> getCommentsByMap(Model model, AuthenticationToken token) throws UserAccessException;
	
	List<Comment> getCommentsByProject(Project project, AuthenticationToken token) throws UserAccessException;

	/**
	 * Returns number of comments in the system.
	 * 
	 * @return number of comments in the system
	 */
	long getCommentCount();

	/**
	 * Removes comments from the model.
	 * 
	 * @param model
	 *          from which model we want to remove comments
	 */
	void removeCommentsForModel(Model model);

	/**
	 * This method returns marker for all comments in a model.
	 * 
	 * @param model
	 *          object to which we are looking for comments
	 * @return list of markers for comments
	 */
	List<FullCommentView> getCommentMarkers(Model model);

	/**
	 * Removes comments from the model.
	 * 
	 * @param model
	 *          from which model we want to remove comments
	 */
	void removeCommentsForModel(ModelData model);

	/**
	 * Returns detailed comment information about element.
	 * 
	 * @param element
	 *          element for which we are looking for comments
	 * @param model
	 *          model where we are looking for comments
	 * @return detailed comment information about element
	 */
	List<ElementIdentifierDetails> getElementInformationForResult(ElementIdentifier element, Model model);

}
