package lcsb.mapviewer.services.search;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;

import lcsb.mapviewer.annotation.data.TargetType;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.model.map.AnnotatedObject;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.species.Complex;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Gene;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.map.species.Rna;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.services.search.db.GeneRow;
import lcsb.mapviewer.services.search.db.TargetView;
import lcsb.mapviewer.services.view.AnnotationViewFactory;

public class ElementMatcher {
	
	/**
	 * Default class logger.
	 */
	private final Logger logger = Logger.getLogger(ElementMatcher.class);

	/**
	 * Factory object used for creation of
	 * {@link lcsb.mapviewer.services.view.AnnotationView} elements.
	 */
	@Autowired
	private AnnotationViewFactory annotationViewFactory;
	
	
	/**
	 * Check if the target should be applied to the element.
	 * 
	 * @param target
	 *          drug target to check
	 * @param element
	 *          {@link Element} to check
	 * @return <code>true</code> if drug target can target element,
	 *         <code>false</code> otherwise
	 */
	public boolean elementMatch(TargetView target, AnnotatedObject element) {
		if (target.getType().equals(TargetType.COMPLEX_PROTEIN.getCommonName())) {
			MiriamData targetAnnotation = annotationViewFactory.viewToObject(target.getAnnotation());
			if (MiriamType.CHEMBL_TARGET.equals(targetAnnotation.getDataType())) {
				if (element.getMiriamData().contains(targetAnnotation)) {
					return true;
				}
			}
			Collection<String> ids = new HashSet<String>();
			for (GeneRow row : target.getProteins()) {
				ids.add(row.getAnnotation().getName());
			}

			if (element instanceof Complex) {
				Complex complex = (Complex) element;
				Set<Species> speciesSet = complex.getAllSimpleChildren();
				if (speciesSet.size() != ids.size()) {
					return false;
				}

				for (Element id : speciesSet) {
					if (!ids.contains(id.getName())) {
						return false;
					}
				}
				return true;
			}
			return false;
		} else if (target.getType().equals(TargetType.SINGLE_PROTEIN.getCommonName())) {
			MiriamData targetAnnotation = null;
			if (target.getAnnotation() != null) {
				targetAnnotation = annotationViewFactory.viewToObject(target.getAnnotation());
			} else if (target.getProteins().size() > 0) {
				targetAnnotation = annotationViewFactory.viewToObject(target.getProteins().get(0).getAnnotation());
			} else {
				logger.warn("Invalid target found: " + target);
			}
			if (targetAnnotation != null && element.getMiriamData().contains(targetAnnotation)) {
				return true;
			}
			if (element instanceof Protein || element instanceof Rna || element instanceof Gene) {
				if (target.getProteins().size() > 0) {
					String hgncId = target.getProteins().get(0).getAnnotation().getName();
					if (element.getName().equalsIgnoreCase(hgncId)) {
						return true;
					}
				}
			}
			return false;
		} else if (target.getType().equals(TargetType.PROTEIN_FAMILY.getCommonName())) {
			if (element instanceof Protein || element instanceof Rna || element instanceof Gene) {
				String hgncId = target.getProteins().get(0).getAnnotation().getName();
				if (element.getName().equalsIgnoreCase(hgncId)) {
					return true;
				} else if (element.getMiriamData().contains(new MiriamData(MiriamType.HGNC_SYMBOL, hgncId))) {
					return true;
				}
				for (GeneRow protein : target.getProteins()) {
					hgncId = protein.getAnnotation().getName();
					if (element.getName().equalsIgnoreCase(hgncId)) {
						return true;
					} else if (element.getMiriamData().contains(new MiriamData(MiriamType.HGNC_SYMBOL, hgncId))) {
						return true;
					}
				}
			}
			return false;
		} else if (target.getType().equals(TargetType.OTHER.getCommonName())) {
			// in other case just compare names
			String name = target.getProteins().get(0).getAnnotation().getName();
			return element.getName().equalsIgnoreCase(name);
		} else {
			throw new InvalidArgumentException("Unknown drug target type: " + target.getType());
		}
	}


	/**
	 * @return the annotationViewFactory
	 * @see #annotationViewFactory
	 */
	public AnnotationViewFactory getAnnotationViewFactory() {
		return annotationViewFactory;
	}


	/**
	 * @param annotationViewFactory the annotationViewFactory to set
	 * @see #annotationViewFactory
	 */
	public void setAnnotationViewFactory(AnnotationViewFactory annotationViewFactory) {
		this.annotationViewFactory = annotationViewFactory;
	}

}
