package lcsb.mapviewer.services.search;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.annotation.data.Article;
import lcsb.mapviewer.annotation.data.TargetType;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Complex;
import lcsb.mapviewer.model.map.species.Gene;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.map.species.Rna;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.services.search.data.ElementIdentifier;
import lcsb.mapviewer.services.search.data.ElementIdentifier.ElementIdentifierType;
import lcsb.mapviewer.services.search.db.GeneRow;
import lcsb.mapviewer.services.search.db.TargetDetails;
import lcsb.mapviewer.services.search.db.TargetView;
import lcsb.mapviewer.services.view.AnnotationViewFactory;

/**
 * Factory for {@link ISearchResultView search results}.
 * 
 * @author Piotr Gawron
 *
 * @param <T>
 *          type from which objects arec reated
 * @param <S>
 *          search ruslt type
 */
@Transactional(value = "txManager")
public abstract class SearchResultFactory<T, S extends ISearchResultView> extends ElementViewFactory<T, S> {

	/**
	 * Default class logger.
	 */
	private final Logger logger = Logger.getLogger(SearchResultFactory.class);

	/**
	 * Creates list of object identifiers from search result.
	 * 
	 * @param object
	 *          search result from which elements should be extracted
	 * @param model
	 *          model on which search was performed
	 * @return list of lightweighted structures for given search result
	 */
	public abstract List<ElementIdentifier> searchResultToElementIdentifier(S object, Model model);

	/**
	 * Factory object used for creation of
	 * {@link lcsb.mapviewer.services.view.AnnotationView} elements.
	 */
	@Autowired
	private AnnotationViewFactory annotationViewFactory;

	@Autowired
	private ElementMatcher elementMatcher;

	/**
	 * Returns detailed information about element respective to the set of targets
	 * (which should be connected to single drug/mirna/chemical).
	 * 
	 * @param element
	 *          element for which we are looking for detailed information
	 * @param model
	 *          {@link Model} on which element is located
	 * @param targets
	 *          list of targets from which detailed information should be
	 *          extracted
	 * @return detailed information for element in the specific drug,
	 *         <code>null</code> if such information doesnt exist
	 */
	public TargetDetails getElementInformationForResult(Collection<TargetView> targets, ElementIdentifier element, Model model) {
		if (ElementIdentifierType.REACTION.getJsName().equalsIgnoreCase(element.getType())) {
			logger.warn("Detail information for reactions not available");
			return null;
		} else if (ElementIdentifierType.POINT.getJsName().equalsIgnoreCase(element.getType())) {
			logger.warn("Detail information for points not available");
			return null;
		} else if (ElementIdentifierType.ALIAS.getJsName().equalsIgnoreCase(element.getType())) {
			Model m = model.getSubmodelById(element.getModelId());
			Element alias = m.getElementByDbId(Integer.valueOf(element.getObjectId()));
			if (alias == null) {
				logger.warn("Alias with object id = " + element.getObjectId() + " is unknown.");
				return null;
			}
			for (TargetView row : targets) {
				if (elementMatch(row, alias)) {
					TargetDetails result = new TargetDetails();
					result.setDatabaseSource(row.getAnnotation());
					for (Article article : row.getReferences()) {
						result.getReferences().add(annotationViewFactory.create(article));
					}
					for (GeneRow protein : row.getProteins()) {
						result.getReferences().add(protein.getAnnotation());
					}
					return result;
				}
			}
		} else {
			throw new InvalidArgumentException("Unknown element type: " + element.getType());
		}
		return null;
	}


	protected boolean elementMatch(TargetView target, Element element) {
		return elementMatcher.elementMatch(target, element);
	}


	/**
	 * @return the annotationViewFactory
	 * @see #annotationViewFactory
	 */
	public AnnotationViewFactory getAnnotationViewFactory() {
		return annotationViewFactory;
	}

	/**
	 * @param annotationViewFactory
	 *          the annotationViewFactory to set
	 * @see #annotationViewFactory
	 */
	public void setAnnotationViewFactory(AnnotationViewFactory annotationViewFactory) {
		this.annotationViewFactory = annotationViewFactory;
	}


	/**
	 * @return the elementMatcher
	 * @see #elementMatcher
	 */
	public ElementMatcher getElementMatcher() {
		return elementMatcher;
	}


	/**
	 * @param elementMatcher the elementMatcher to set
	 * @see #elementMatcher
	 */
	public void setElementMatcher(ElementMatcher elementMatcher) {
		this.elementMatcher = elementMatcher;
	}

}
