package lcsb.mapviewer.services.search.data;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.apache.log4j.Logger;
import org.primefaces.model.DefaultTreeNode;
import org.primefaces.model.TreeNode;
import org.springframework.beans.factory.annotation.Autowired;

import com.google.gson.Gson;

import lcsb.mapviewer.annotation.data.Chebi;
import lcsb.mapviewer.annotation.data.ChebiRelation;
import lcsb.mapviewer.annotation.services.MiriamConnector;
import lcsb.mapviewer.annotation.services.annotators.ChebiAnnotator;
import lcsb.mapviewer.annotation.services.annotators.ChebiSearchException;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.common.exception.NotImplementedException;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamRelationType;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.graph.DataMining;
import lcsb.mapviewer.model.map.graph.DataMiningSet;
import lcsb.mapviewer.model.map.graph.DataMiningType;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.model.map.species.field.ModificationResidue;
import lcsb.mapviewer.services.interfaces.IDataMiningService;
import lcsb.mapviewer.services.overlay.ChebiTreeRow;
import lcsb.mapviewer.services.search.ElementViewFactory;
import lcsb.mapviewer.services.search.data.FullAliasView.PosttransationalModification;
import lcsb.mapviewer.services.view.AnnotationView;
import lcsb.mapviewer.services.view.AnnotationViewFactory;
import lcsb.mapviewer.services.view.DataMiningView;
import lcsb.mapviewer.services.view.DataMiningViewFactory;

/**
 * Factory class for {@link LightAliasView} class.
 * 
 * @author Piotr Gawron
 * 
 */
public class FullAliasViewFactory extends ElementViewFactory<Element, FullAliasView> {

	/**
	 * How many rows with data mining information should be presented to the user.
	 */
	private static final int			MAX_MISSING_CONNECTIONS	= 100;

	/**
	 * Default class logger.
	 */
	private static Logger					logger									= Logger.getLogger(FullAliasViewFactory.class);

	/**
	 * Factory used to create {@link AnnotationView} elements.
	 */
	@Autowired
	private AnnotationViewFactory	annotationViewFactory;

	/**
	 * Connector used for accessing data from miriam registry.
	 */
	@Autowired
	private MiriamConnector				mc;

	/**
	 * Factory object used for creation of {@link DataMiningView} elements.
	 */
	@Autowired
	private DataMiningViewFactory	dataMiningViewFactory;

	/**
	 * Service that manages data mining information.
	 */
	@Autowired
	private IDataMiningService		dataMiningService;

	/**
	 * Service accessing chebi database.
	 */
	@Autowired
	private ChebiAnnotator				chebiBackend;

	@Override
	public FullAliasView create(Element alias) {
		return create(alias, null);
	}

	/**
	 * Creates {@link FullAliasView} from given {@link Element}.
	 * 
	 * @param element
	 *          alias for which view is created
	 * @param icon
	 *          icon that should be assigned to the view
	 * @return {@link FullAliasView} for given alias
	 */
	public FullAliasView create(Element element, String icon) {
		FullAliasView result = new FullAliasView(element, icon);
		if (element != null) {
			result.addReferences(annotationViewFactory.createList(element.getMiriamData()));
			logger.debug("datamining start");
			result.addOther("dataMining", getDataMiningInformation(element, MAX_MISSING_CONNECTIONS));
			logger.debug("datamining stop");

			List<AnnotationView> columns = new ArrayList<>();
			AnnotationView structureView = annotationViewFactory.create(DataMiningType.STRUCTURE_ANALYSIS);

			AnnotationView textMiningView = annotationViewFactory.create(DataMiningType.TEXT_MINING);

			columns.add(structureView);
			columns.add(textMiningView);

			result.addOther("dataMiningColumns", columns);

			if (element.getModel() != null) {
				for (DataMiningSet dms : element.getModel().getDataMiningSets()) {
					if (DataMiningType.STRUCTURE_ANALYSIS.equals(dms.getType())) {
						if (dms.getSource() != null && !dms.getSource().isEmpty()) {
							structureView.setLink(dms.getSource());
						}
						if (dms.getName() != null && !dms.getName().isEmpty()) {
							structureView.setName(dms.getName());
						}
						structureView.setSummary(dms.getDescription());

					} else if (DataMiningType.TEXT_MINING.equals(dms.getType())) {
						if (dms.getSource() != null && !dms.getSource().isEmpty()) {
							textMiningView.setLink(dms.getSource());
						}
						if (dms.getName() != null && !dms.getName().isEmpty()) {
							textMiningView.setName(dms.getName());
						}
						textMiningView.setSummary(dms.getDescription());
					} else {
						throw new NotImplementedException(dms.getType() + " is not implemented.");
					}
				}
			}
			for (MiriamData md : element.getMiriamData()) {
				if (MiriamType.CHEBI.equals(md.getDataType())) {
					try {
						Chebi chebi = chebiBackend.getChebiElementForChebiId(md);
						if (chebi != null) {
							result.addOther("chebiTree", chebi);
						} else {
							logger.warn("Invalid chebi identifier: " + md.getResource());
						}
					} catch (ChebiSearchException e) {
						logger.warn("Problem with accesing data about chebi: " + md.getResource(), e);
					}
				}
			}
			if (element instanceof Protein) {
				result.addOther("posttranslationalModifications", getPosttransationalModifications(element));
			}

		}
		return result;
	}

	/**
	 * Creates list of {@link PosttransationalModification} objects for object
	 * alias.
	 * 
	 * @param alias
	 *          alias of the object
	 * @return list of {@link PosttransationalModification} objects for object
	 *         alias
	 */
	private List<PosttransationalModification> getPosttransationalModifications(Element alias) {
		List<PosttransationalModification> result = new ArrayList<PosttransationalModification>();
		if (alias instanceof Protein) {
			List<ModificationResidue> modifications = ((Protein) alias).getModificationResidues();
			if (modifications.size() > 0) {
				for (ModificationResidue residue : modifications) {
					if (residue.getState() != null) {
						PosttransationalModification modification = new PosttransationalModification();
						modification.setName(residue.getState().getFullName());
						modification.setPosition(residue.getName());
						result.add(modification);
					}
				}
			}
		}
		return result;
	}

	@Override
	public String createGson(FullAliasView object) {
		return new Gson().toJson(object);
	}

	/**
	 * Creates table for data mining information for given alias.
	 * 
	 * @param alias
	 *          alias for which data mining information is added
	 * @param maxMissingConnections
	 *          max elements that should be presented
	 * @return html table with data mining information
	 */
	protected List<List<DataMiningView>> getDataMiningInformation(Element alias, int maxMissingConnections) {
		List<List<DataMiningView>> result = new ArrayList<>();
		List<DataMiningView> enricoList = new ArrayList<>();
		List<DataMiningView> mariaList = new ArrayList<>();
		if (!(alias instanceof Species)) {
			return result;
		}
		if (dataMiningService != null) {
			logger.debug("X");
			Collection<DataMining> connections = alias.getDataMining();
			logger.debug("Y");
			if (connections.size() > 0) {
				for (DataMining missingConnection : connections) {
					if (missingConnection.getType().equals(DataMiningType.STRUCTURE_ANALYSIS)) {
						if (enricoList.size() < maxMissingConnections) {
							DataMiningView view = dataMiningViewFactory.create(missingConnection);
							enricoList.add(view);
						}
					} else if (missingConnection.getType().equals(DataMiningType.TEXT_MINING)) {
						if (mariaList.size() < maxMissingConnections) {
							DataMiningView view = dataMiningViewFactory.create(missingConnection);
							mariaList.add(view);
						}
					} else {
						throw new InvalidArgumentException("Unknown type of missing connection: " + missingConnection.getType());
					}
				}
			}
		}
		int size = Math.max(mariaList.size(), enricoList.size());
		for (int i = 0; i < size; i++) {
			List<DataMiningView> row = new ArrayList<DataMiningView>();
			if (i < enricoList.size()) {
				row.add(enricoList.get(i));
			} else {
				row.add(null);
			}
			if (i < mariaList.size()) {
				row.add(mariaList.get(i));
			} else {
				row.add(null);
			}
			result.add(row);
		}
		return result;
	}

	/**
	 * Creates empty {@link FullAliasView}.
	 * 
	 * @return empty {@link FullAliasView}
	 */
	public FullAliasView createEmpty() {
		return new FullAliasView();
	}

	/**
	 * Creates {@link TreeNode} representing {@link Chebi} object. This object is
	 * used by client side to represent data as a tree.
	 * 
	 * @param chebi
	 *          object to be transforem
	 * @return tree representing chebi element
	 */

	protected TreeNode createTreeForChebi(Chebi chebi) {
		// create artificial root

		TreeNode root = new DefaultTreeNode(new ChebiTreeRow("root", "unknown", MiriamType.CHEBI.getDbHomepage(), ""), null);
		root.setExpanded(true);
		TreeNode lastElement = root;
		// create parents
		for (ChebiRelation relation : chebi.getIncomingChebi()) {
			MiriamData md = new MiriamData(MiriamRelationType.BQ_BIOL_IS_DESCRIBED_BY, MiriamType.CHEBI, relation.getId());
			ChebiTreeRow row = new ChebiTreeRow(relation.getName(), relation.getId(), mc.getUrlString(md), relation.getType());
			lastElement = new DefaultTreeNode(row, root);
			lastElement.setExpanded(true);
		}

		// create main chebi element
		MiriamData md = new MiriamData(MiriamRelationType.BQ_BIOL_IS_DESCRIBED_BY, MiriamType.CHEBI, chebi.getChebiId());
		ChebiTreeRow row = new ChebiTreeRow(chebi.getName(), chebi.getChebiId(), mc.getUrlString(md), "");

		TreeNode element = new DefaultTreeNode(row, lastElement);
		element.setExpanded(true);

		// create children

		for (ChebiRelation relation : chebi.getOutgoingChebi()) {
			md = new MiriamData(MiriamRelationType.BQ_BIOL_IS_DESCRIBED_BY, MiriamType.CHEBI, relation.getId());
			row = new ChebiTreeRow(relation.getName(), relation.getId(), mc.getUrlString(md), relation.getType());
			new DefaultTreeNode(row, element);
		}

		return root;
	}

}
