package lcsb.mapviewer.services.view;

import java.io.Serializable;
import java.util.List;

import lcsb.mapviewer.model.map.model.Model;

import org.apache.log4j.Logger;
import org.primefaces.model.map.LatLng;

/**
 * View representatin of a model. It's a light structre that present
 * {@link Model} data to the client side.
 * 
 * @author Piotr Gawron
 * 
 */
public class ModelView extends AbstractView<Model> implements Serializable {
	/**
	 * Default class logger.
	 */
	@SuppressWarnings("unused")
	private static Logger						logger						= Logger.getLogger(ModelView.class);
	/**
	 * 
	 */
	private static final long				serialVersionUID	= 1L;

	/**
	 * Version of the model.
	 */
	private String									version;

	/**
	 * Name of the model.
	 */
	private String									name;

	/**
	 * {@link lcsb.mapviewer.model.Project#projectId Project identifier}.
	 */
	private String									projectId;

	/**
	 * Description of the model.
	 */
	private String									description;
	/**
	 * Size in pixels of the single square tile (small image used for graphical
	 * representation).
	 */
	private Integer									tileSize;
	/**
	 * Size of the whole map.
	 */
	private Integer									pictureSize;
	/**
	 * Minimum zoom level that should be allowed by the Google Maps API.
	 */
	private Integer									minZoom;
	/**
	 * Maximum zoom level that should be allowed by the Google Maps API.
	 */
	private Integer									maxZoom;
	/**
	 * Public layouts available for the model.
	 */
	private List<LayoutView>				layouts;

	/**
	 * Custom layouts available only for the current user.
	 */
	private List<LayoutView>				customLayouts;

	/**
	 * List of submodels in the model.
	 */
	private List<ModelView>					submodels;

	/**
	 * Where is the center of the map in latituted, longiude format.
	 */
	private LatLng									centerLatLng;

	/**
	 * Top-Left corner of the map (0,0) as a latLng coordinates.
	 */
	private LatLng									topLeftLatLng;

	/**
	 * Bottom-Right corner of the map (width,height) as a latLng coordinates.
	 */
	private LatLng									bottomRightLatLng;

	/**
	 * Should the map be automatically fit in the browser window. Bean should
	 * change this value if the map was already viewed by user and there is
	 * information about zoom level and coordinates where user is browsing it.
	 */
	private Boolean									fitMapBounds			= true;

	/**
	 * List of overview images attached to this model.
	 */
	private List<OverviewImageView>	overviewImageViews;

	/**
	 * Top level overview image.
	 */
	private OverviewImageView				topOverviewImage;

	/**
	 * Default constructor which creates view representation from the orignial
	 * model.
	 * 
	 * @param model
	 *          source map model
	 */
	protected ModelView(Model model) {
		super(model);
	}

	/**
	 * Default constructor. Should be used only for deserialization.
	 */
	protected ModelView() {
	}

	/**
	 * @return the version
	 * @see #version
	 */
	public String getVersion() {
		return version;
	}

	/**
	 * @param version
	 *          the version to set
	 * @see #version
	 */
	public void setVersion(String version) {
		this.version = version;
	}

	/**
	 * @return the description
	 * @see #description
	 */
	public String getDescription() {
		return description;
	}

	/**
	 * @param description
	 *          the description to set
	 * @see #description
	 */
	public void setDescription(String description) {
		this.description = description;
	}

	/**
	 * @return the tileSize
	 * @see #tileSize
	 */
	public Integer getTileSize() {
		return tileSize;
	}

	/**
	 * @param tileSize
	 *          the tileSize to set
	 * @see #tileSize
	 */
	public void setTileSize(Integer tileSize) {
		this.tileSize = tileSize;
	}

	/**
	 * @return the pictureSize
	 * @see #pictureSize
	 */
	public Integer getPictureSize() {
		return pictureSize;
	}

	/**
	 * @param pictureSize
	 *          the pictureSize to set
	 * @see #pictureSize
	 */
	public void setPictureSize(Integer pictureSize) {
		this.pictureSize = pictureSize;
	}

	/**
	 * @return the minZoom
	 * @see #minZoom
	 */
	public Integer getMinZoom() {
		return minZoom;
	}

	/**
	 * @param minZoom
	 *          the minZoom to set
	 * @see #minZoom
	 */
	public void setMinZoom(Integer minZoom) {
		this.minZoom = minZoom;
	}

	/**
	 * @return the maxZoom
	 * @see #maxZoom
	 */
	public Integer getMaxZoom() {
		return maxZoom;
	}

	/**
	 * @param maxZoom
	 *          the maxZoom to set
	 * @see #maxZoom
	 */
	public void setMaxZoom(Integer maxZoom) {
		this.maxZoom = maxZoom;
	}

	/**
	 * @return the layouts
	 * @see #layouts
	 */
	public List<LayoutView> getLayouts() {
		return layouts;
	}

	/**
	 * @param layouts
	 *          the layouts to set
	 * @see #layouts
	 */
	public void setLayouts(List<LayoutView> layouts) {
		this.layouts = layouts;
	}

	/**
	 * @return the centerLatLng
	 * @see #centerLatLng
	 */
	public LatLng getCenterLatLng() {
		return centerLatLng;
	}

	/**
	 * @param centerLatLng
	 *          the centerLatLng to set
	 * @see #centerLatLng
	 */
	public void setCenterLatLng(LatLng centerLatLng) {
		this.centerLatLng = centerLatLng;
	}

	/**
	 * @return the name
	 * @see #name
	 */
	public String getName() {
		return name;
	}

	/**
	 * @param name
	 *          the name to set
	 * @see #name
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 * @return the customLayouts
	 * @see #customLayouts
	 */
	public List<LayoutView> getCustomLayouts() {
		return customLayouts;
	}

	/**
	 * @param customLayouts
	 *          the customLayouts to set
	 * @see #customLayouts
	 */
	public void setCustomLayouts(List<LayoutView> customLayouts) {
		this.customLayouts = customLayouts;
	}

	/**
	 * @return the submodels
	 * @see #submodels
	 */
	public List<ModelView> getSubmodels() {
		return submodels;
	}

	/**
	 * @param submodels
	 *          the submodels to set
	 * @see #submodels
	 */
	public void setSubmodels(List<ModelView> submodels) {
		this.submodels = submodels;
	}

	/**
	 * @return the overviewImageViews
	 * @see #overviewImageViews
	 */
	public List<OverviewImageView> getOverviewImageViews() {
		return overviewImageViews;
	}

	/**
	 * @param overviewImageViews
	 *          the overviewImageViews to set
	 * @see #overviewImageViews
	 */
	public void setOverviewImageViews(List<OverviewImageView> overviewImageViews) {
		this.overviewImageViews = overviewImageViews;
	}

	/**
	 * @return the topOverviewImage
	 * @see #topOverviewImage
	 */
	public OverviewImageView getTopOverviewImage() {
		return topOverviewImage;
	}

	/**
	 * @param topOverviewImage
	 *          the topOverviewImage to set
	 * @see #topOverviewImage
	 */
	public void setTopOverviewImage(OverviewImageView topOverviewImage) {
		this.topOverviewImage = topOverviewImage;
	}

	/**
	 * @return the topLeftLatLng
	 * @see #topLeftLatLng
	 */
	public LatLng getTopLeftLatLng() {
		return topLeftLatLng;
	}

	/**
	 * @param topLeftLatLng
	 *          the topLeftLatLng to set
	 * @see #topLeftLatLng
	 */
	public void setTopLeftLatLng(LatLng topLeftLatLng) {
		this.topLeftLatLng = topLeftLatLng;
	}

	/**
	 * @return the bottomRightLatLng
	 * @see #bottomRightLatLng
	 */
	public LatLng getBottomRightLatLng() {
		return bottomRightLatLng;
	}

	/**
	 * @param bottomRightLatLng
	 *          the bottomRightLatLng to set
	 * @see #bottomRightLatLng
	 */
	public void setBottomRightLatLng(LatLng bottomRightLatLng) {
		this.bottomRightLatLng = bottomRightLatLng;
	}

	/**
	 * @return the fitMapBounds
	 * @see #fitMapBounds
	 */
	public Boolean getFitMapBounds() {
		return fitMapBounds;
	}

	/**
	 * @param fitMapBounds
	 *          the fitMapBounds to set
	 * @see #fitMapBounds
	 */
	public void setFitMapBounds(Boolean fitMapBounds) {
		this.fitMapBounds = fitMapBounds;
	}

	/**
	 * @return the projectId
	 * @see #projectId
	 */
	public String getProjectId() {
		return projectId;
	}

	/**
	 * @param projectId
	 *          the projectName to set
	 * @see #projectId
	 */
	public void setProjectId(String projectId) {
		this.projectId = projectId;
	}

}
