package lcsb.mapviewer.services.view;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;

import lcsb.mapviewer.common.Configuration;
import lcsb.mapviewer.model.map.layout.Layout;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelSubmodelConnection;
import lcsb.mapviewer.services.utils.gmap.CoordinationConverter;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.common.exception.NotImplementedException;

import com.google.gson.Gson;

/**
 * Factory class for {@link ModelView} class.
 * 
 * @author Piotr Gawron
 * 
 */
@Transactional(value = "txManager")
public class ModelViewFactory extends AbstractViewFactory<Model, ModelView> {
	/**
	 * Default class logger.
	 */
	private static Logger							logger	= Logger.getLogger(ModelViewFactory.class);

	/**
	 * Factory object used for creation of {@link LayoutView} elements.
	 */
	@Autowired
	private LayoutViewFactory					layoutViewFactory;

	/**
	 * Factory object used for creation of {@link OverviewImageView} elements.
	 */
	@Autowired
	private OverviewImageViewFactory	overviewImageViewFactory;

	@Override
	public ModelView create(Model model) {
		ModelView result = new ModelView(model);
		if (model == null) {
			logger.warn("Empty model...");
			return result;
		}
		if (model.getProject() != null) {
			result.setProjectId(model.getProject().getProjectId());
		}
		result.setLayouts(new ArrayList<LayoutView>());
		result.setCustomLayouts(new ArrayList<LayoutView>());
		result.setVersion(model.getMapVersion());
		result.setName(model.getName());

		result.setMinZoom(Configuration.MIN_ZOOM_LEVEL);
		result.setMaxZoom(result.getMinZoom() + model.getZoomLevels());
		result.setTileSize(model.getTileSize());
		result.setPictureSize((int) Math.max(model.getWidth(), model.getHeight()));
		// we have to add copies (originals contain reference to model which
		// shouldn't be passed to simple structure)
		for (Layout layout : model.getLayouts()) {
			if (layout.isPublicLayout()) {
				result.getLayouts().add(layoutViewFactory.create(layout));
			}
		}
		CoordinationConverter cConverter = new CoordinationConverter(model);
		result.setCenterLatLng(cConverter.toLatLng(new Point2D.Double(result.getPictureSize() / 2, result.getPictureSize() / 2)));
		result.setBottomRightLatLng(cConverter.toLatLng(new Point2D.Double(model.getWidth(), model.getHeight())));
		result.setTopLeftLatLng(cConverter.toLatLng(new Point2D.Double(0, 0)));

		result.setDescription(model.getNotes());

		List<ModelView> submodels = new ArrayList<ModelView>();
		for (ModelSubmodelConnection connection : model.getSubmodelConnections()) {
			submodels.add(create(connection.getSubmodel().getModel()));
		}
		result.setSubmodels(submodels);
		result.setOverviewImageViews(overviewImageViewFactory.createList(model.getOverviewImages()));
		if (result.getOverviewImageViews().size() > 0) {
			result.setTopOverviewImage(overviewImageViewFactory.create(model.getTopOverviewImage()));
		}
		return result;
	}

	@Override
	public String createGson(ModelView object) {
		return new Gson().toJson(object);
	}

	@Override
	public Model viewToObject(ModelView view) {
		throw new NotImplementedException();
	}
}
