package lcsb.mapviewer.services.view;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;

import com.google.gson.Gson;

import lcsb.mapviewer.annotation.data.MeSH;
import lcsb.mapviewer.annotation.services.MeSHParser;
import lcsb.mapviewer.annotation.services.annotators.AnnotatorException;
import lcsb.mapviewer.common.exception.NotImplementedException;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.map.graph.DataMiningSet;
import lcsb.mapviewer.model.map.layout.Layout;
import lcsb.mapviewer.model.map.model.ModelData;

/**
 * Factory class for {@link ProjectView} class.
 * 
 * @author Piotr Gawron
 * 
 */
public class ProjectViewFactory extends AbstractViewFactory<Project, ProjectView> {

	/**
	 * Default class logger.
	 */
	private static Logger					logger = Logger.getLogger(ProjectViewFactory.class);

	/**
	 * Factory object for {@link LayoutView} elements.
	 */
	@Autowired
	private LayoutViewFactory			layoutViewFactory;

	/**
	 * Factory object used for creation of {@link AnnotationView} elements.
	 */
	@Autowired
	private AnnotationViewFactory	annotationViewFactory;

	/**
	 * Access object for information about mesh terms.
	 */
	@Autowired
	private MeSHParser						meSHParser;

	@Override
	public ProjectView create(Project project) {
		ProjectView result = new ProjectView(project);
		if (project == null) {
			return result;
		}
		result.setIdObject(project.getId());
		result.setErrors(project.getErrors());
		result.setProjectId(project.getProjectId());
		result.setProjectName(project.getName());
		if (project.getDisease() != null) {
			result.setProjectDiseaseLink(annotationViewFactory.create(project.getDisease()));
			result.setNewDiseaseName(project.getDisease().getResource());
			MeSH mesh;
			try {
				mesh = meSHParser.getMeSH(project.getDisease());
				if (mesh != null) {
					result.setDiseaseName(mesh.getName());
				}
			} catch (AnnotatorException e) {
				logger.error("Problem with getting info about mesh entry: ", e);
			}
		}
		
		if (project.getOrganism() != null) {
			result.setOrganismLink(annotationViewFactory.create(project.getOrganism()));
			result.setNewOrganismName(project.getOrganism().getResource());
		}
		result.setStatus(project.getStatus().toString());
		result.setProgress(Integer.valueOf((int) project.getProgress()));
		result.setInputDataAvailable(project.getInputFileName() != null);

		int warnCount = 0;
		String warnPreview = "";
		for (String w : project.getWarnings()) {
			if (!"".equals(w)) {
				if (warnCount <= 2) {
					warnPreview += w + "\n";
				}
				warnCount++;
			}
		}
		result.setWarningsPreview(warnPreview);
		result.setWarnings(project.getWarnings());

		for (ModelData model : project.getModels()) {
			result.setModelId(model.getId());
			result.setDescription(model.getNotes());
			result.getLayouts().clear();
			result.setVersion(model.getMapVersion());
			result.setNotifyEmail(model.getNotifyEmail());
			for (Layout layout : model.getLayouts()) {
				LayoutView layoutRow = layoutViewFactory.create(layout);
				result.addLayout(layoutRow);
			}
			for (DataMiningSet dms : model.getDataMiningSets()) {
				result.addDataMiningSet(new DataMiningSetView(dms));
			}
		}
		return result;
	}

	@Override
	public String createGson(ProjectView object) {
		return new Gson().toJson(object);
	}

	@Override
	public Project viewToObject(ProjectView view) {
		throw new NotImplementedException();
	}
}
