package lcsb.mapviewer.services.view;

import java.util.ArrayList;
import java.util.List;

import lcsb.mapviewer.annotation.data.Article;
import lcsb.mapviewer.common.Pair;
import lcsb.mapviewer.common.exception.InvalidStateException;
import lcsb.mapviewer.model.map.AnnotatedObject;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.services.impl.SearchService;

/**
 * View object for pubmed article for a given model. It contains information
 * about article and list of elements annotated by this
 * {@link lcsb.mapviewer.model.map.MiriamType#PUBMED article}.
 * 
 * @author Piotr Gawron
 *
 */
public class PubmedAnnotatedElementsView extends AbstractView<MiriamData> {

	/**
	 * 
	 */
	private static final long					 serialVersionUID	= 1L;

	/**
	 * Article for which this pubmed summary is created.
	 */
	private Article										 article;

	/**
	 * List of elements that are annotated the publication. {@link Pair#left}
	 * refers to human readable name and {@link Pair#right} is search string that
	 * can be used to access element on the model.
	 */
	private List<Pair<String, String>> elements					= new ArrayList<>();

	/**
	 * Default constructor.
	 * 
	 * @param element
	 *          {@link MiriamData} that represents publication for this object.
	 * 
	 * @see #article
	 */
	public PubmedAnnotatedElementsView(MiriamData element) {
		super(element);
	}

	/**
	 * @return the article
	 * @see #article
	 */
	public Article getArticle() {
		return article;
	}

	/**
	 * @param article
	 *          the article to set
	 * @see #article
	 */
	public void setArticle(Article article) {
		this.article = article;
	}

	/**
	 * @return the elements
	 * @see #elements
	 */
	public List<Pair<String, String>> getElements() {
		return elements;
	}

	/**
	 * @param elements
	 *          the elements to set
	 * @see #elements
	 */
	public void setElements(List<Pair<String, String>> elements) {
		this.elements = elements;
	}

	/**
	 * Adds element annotratd by {@link #article}.
	 * 
	 * @param annotatedObject
	 *          object to add
	 */
	public void addElement(AnnotatedObject annotatedObject) {
		if (annotatedObject instanceof Element) {
			Element alias = (Element) annotatedObject;
			elements.add(new Pair<String, String>(alias.getName(), SearchService.SPECIES_SEARCH_PREFIX + ":" + alias.getElementId()));
		} else if (annotatedObject instanceof Reaction) {
			Reaction reaction = (Reaction) annotatedObject;
			elements.add(
					new Pair<>(
							SearchService.REACTION_SEARCH_PREFIX + ":" + reaction.getIdReaction(), SearchService.REACTION_SEARCH_PREFIX + ":" + reaction.getIdReaction()));
		} else {
			throw new InvalidStateException("Unknown subtype of " + AnnotatedObject.class.getName() + ": " + annotatedObject.getClass().getName());
		}

	}
}
