package lcsb.mapviewer.services.search.db.drug;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import lcsb.mapviewer.annotation.data.Drug;
import lcsb.mapviewer.annotation.services.TaxonomyBackend;
import lcsb.mapviewer.converter.ComplexZipConverter;
import lcsb.mapviewer.converter.ComplexZipConverterParams;
import lcsb.mapviewer.converter.model.celldesigner.CellDesignerXmlParser;
import lcsb.mapviewer.converter.zip.ModelZipEntryFile;
import lcsb.mapviewer.converter.zip.ZipEntryFile;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelFullIndexed;
import lcsb.mapviewer.model.map.model.SubmodelType;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.GenericProtein;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.services.ServiceTestFunctions;
import lcsb.mapviewer.services.search.db.DbSearchCriteria;
import lcsb.mapviewer.services.search.db.GeneRow;
import lcsb.mapviewer.services.search.db.TargetView;

public class DrugServiceTest extends ServiceTestFunctions {
	Logger logger = Logger.getLogger(DrugServiceTest.class);

	@Before
	public void setUp() throws Exception {
	}

	@After
	public void tearDown() throws Exception {
	}

	@Test
	public void testGetEmpty() throws Exception {
		try {
			DrugView drug = drugService.getByName("blablablabla", new DbSearchCriteria());
			assertNull(drug);
		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}
	}

	@Test
	public void testGetTargets() throws Exception {
		try {
			long count = searchHistoryDao.getCount();
			Model model = new ModelFullIndexed(null);
			Project project = new Project();
			project.setProjectId("TesT");
			project.addModel(model);
			DrugView drug = drugService.getByName("Diazoxide", new DbSearchCriteria().model(model).ipAddress("ip"));
			assertNotNull(drug);
			assertNotNull(drug.getName());
			assertFalse(drug.getName().trim().equals(""));
			assertNotNull(drug.getDescription());
			assertFalse(drug.getDescription().trim().equals(""));

			String hgnc1 = "SLC12A3";
			String hgnc2 = "KCNJ8";

			boolean hgnc1Exists = false;
			boolean hgnc2Exists = false;

			for (TargetView target : drug.getTargetRows()) {
				for (GeneRow row : target.getProteins()) {
					if (row.getAnnotation().getName().equalsIgnoreCase(hgnc1))
						hgnc1Exists = true;
					if (row.getAnnotation().getName().equalsIgnoreCase(hgnc2))
						hgnc2Exists = true;
				}
			}

			assertTrue("Target " + hgnc1 + " doesn't exist in target list but should", hgnc1Exists);
			assertTrue("Target " + hgnc2 + " doesn't exist in target list but should", hgnc2Exists);

			long count2 = searchHistoryDao.getCount();

			assertEquals(count + 1, count2);

			assertEquals("YES", drug.getBloodBrainBarrier());

		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}
	}

	@Test
	public void testGetTargetsBySynonym() throws Exception {
		try {
			// search by synonym
			DrugView drug = drugService.getByName("Amantidine", new DbSearchCriteria().ipAddress("ip"));
			// search by name
			DrugView drug2 = drugService.getByName("Amantadine", new DbSearchCriteria().ipAddress("ip"));
			assertNotNull(drug);
			assertNotNull(drug.getName());
			assertFalse(drug.getName().trim().equals(""));
			assertNotNull(drug.getDescription());
			assertFalse(drug.getDescription().trim().equals(""));

			// number of targets should be the same
			assertEquals(drug.getTargetRows().size(), drug2.getTargetRows().size());
			assertTrue(drug.getTargetRows().size() > 0);

		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}
	}

	@Test
	public void testFindDrugSelegiline() throws Exception {
		try {
			DrugView test = drugService.getByName("Selegiline", new DbSearchCriteria().ipAddress("ip").organisms(TaxonomyBackend.HUMAN_TAXONOMY));
			assertNotNull(test);
		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}
	}

	@Test
	public void testDornaseAplha() throws Exception {
		try {
			DrugView drug = drugService.getByName("Dornase alpha", new DbSearchCriteria());
			assertNotNull(drug.getName());
			assertEquals("N/A", drug.getBloodBrainBarrier());
		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}

	}

	@Test
	public void testRapamycin() throws Exception {
		try {
			DrugView drug = drugService.getByName("Rapamycin", new DbSearchCriteria());
			assertNotNull(drug.getName());
			assertEquals("NO", drug.getBloodBrainBarrier());
		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}

	}

	@Test
	public void testSearchByElements() throws Exception {
		try {
			List<Element> elements = new ArrayList<>();
			List<DrugView> drugs = drugService.getForTargets(elements, new DbSearchCriteria());
			assertNotNull(drugs);
			assertEquals(0, drugs.size());

			Protein protein = new GenericProtein("id");
			protein.setName("DRD2");
			protein.addMiriamData(new MiriamData(MiriamType.HGNC_SYMBOL, "DRD2"));
			elements.add(protein);

			drugs = drugService.getForTargets(elements, new DbSearchCriteria());
			assertNotNull(drugs);
			assertTrue(drugs.size() > 0);
		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}

	}

	@Test
	public void testSearchByElements2() throws Exception {
		try {
			List<Element> elements = new ArrayList<>();
			Protein protein = new GenericProtein("id");
			protein.setName("DRD2");
			protein.addMiriamData(new MiriamData(MiriamType.HGNC_SYMBOL, "DRD2"));
			elements.add(protein);
			List<DrugView> drugs = drugService.getForTargets(elements, new DbSearchCriteria().organisms(TaxonomyBackend.HUMAN_TAXONOMY));

			assertNotNull(drugs);
			assertTrue(drugs.size() > 0);
		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}

	}

	@Test
	public void testSearchByElements3() throws Exception {
		try {
			List<Element> elements = new ArrayList<>();
			Protein protein = new GenericProtein("id");
			protein.setName("GLUD1");
			protein.addMiriamData(new MiriamData(MiriamType.HGNC_SYMBOL, "GLUD1"));
			protein.addMiriamData(new MiriamData(MiriamType.ENTREZ, "2746"));
			elements.add(protein);
			List<DrugView> drugs = drugService.getForTargets(elements, new DbSearchCriteria());

			assertNotNull(drugs);
			assertTrue(drugs.size() > 0);
		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}

	}

	@Test
	public void testObjectToDrugTargetList() throws Exception {
		try {
			DrugView drug = drugService.getByName("AMANTADINE", new DbSearchCriteria());
			Drug drug2 = drugBankHTMLParser.findDrug("AMANTADINE");

			assertTrue(drug2.getTargets().size() <= drug.getTargetRows().size());

		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}
	}

	@Test
	public void testAspirinToDrugTargetList() throws Exception {
		try {
			DrugView drug = drugService.getByName("Aspirin", new DbSearchCriteria());
			Drug drug2 = drugBankHTMLParser.findDrug("Aspirin");

			assertTrue(drug2.getTargets().size() <= drug.getTargetRows().size());
		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}
	}

	@Test
	public void testIbuprofenMultiDbTargets() throws Exception {
		try {
			Model model = getModelForFile("testFiles/drug_target_2_db.xml", true);
			DrugView drug = drugService.getByName("Ibuprofen", new DbSearchCriteria().model(model));
			Set<String> icons = new HashSet<String>();
			for (TargetView row : drug.getTargetRows()) {
				if (row.getSelectable()) {
					assertFalse("More than one target has icon: " + row.getIcon(), icons.contains(row.getIcon()));
					icons.add(row.getIcon());
				}
			}

		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}
	}

	@Test
	public void testGetDrugByDrugNameInComplex() throws Exception {
		try {
			ComplexZipConverter<CellDesignerXmlParser> parser = new ComplexZipConverter<CellDesignerXmlParser>(CellDesignerXmlParser.class);
			ComplexZipConverterParams complexParams = new ComplexZipConverterParams().zipFile("testFiles/complexModel/complex_drug_target_model.zip");
			ZipEntryFile entry1 = new ModelZipEntryFile("main.xml", "main", true, false, SubmodelType.UNKNOWN);
			ZipEntryFile entry2 = new ModelZipEntryFile("s1.xml", "s1", false, false, SubmodelType.UNKNOWN);
			ZipEntryFile entry3 = new ModelZipEntryFile("s2.xml", "s2", false, false, SubmodelType.UNKNOWN);
			ZipEntryFile entry4 = new ModelZipEntryFile("s3.xml", "s3", false, false, SubmodelType.UNKNOWN);
			ZipEntryFile entry5 = new ModelZipEntryFile("mapping.xml", "mapping", false, true, SubmodelType.UNKNOWN);
			complexParams.entry(entry1);
			complexParams.entry(entry2);
			complexParams.entry(entry3);
			complexParams.entry(entry4);
			complexParams.entry(entry5);
			Model model = parser.createModel(complexParams);

			DrugView drug = drugService.getByName("Aspirin", new DbSearchCriteria().model(model));
			Set<String> icons = new HashSet<String>();

			for (TargetView row : drug.getTargetRows()) {
				if (row.getSelectable()) {
					icons.add(row.getIcon());
				}
			}
			assertTrue("At least 2 targets should be selected", icons.size() >= 2);

		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}
	}

	@Test
	public void testAspirinSynonyms() throws Exception {
		try {
			DrugView drug = drugService.getByName("Aspirin", new DbSearchCriteria());

			Set<String> synonyms = new HashSet<String>();
			for (String string : drug.getSynonyms()) {
				assertFalse("Duplicate entry in drug synonym: " + string, synonyms.contains(string));
				synonyms.add(string);
			}

		} catch (Exception e) {
			e.printStackTrace();
			throw e;
		}
	}

}
